// online-mfcnet.cc
// If you want to put this file under the directory, it file is placed in under the "feat" directory.

#ifdef _MSC_VER
#define NOMINMAX
#endif
#include "online-mfcnet.h"
#include "malloc.h"
namespace kaldi {

//FOR_FILE_MODE//

OnlineFeatureMfcnet::OnlineFeatureMfcnet(
    //kaldi::OnlineFeatureInputSettingOptions &opts,
    kaldi::OnlineNnet2FeaturePipelineInfo *feature_info,
    const kaldi::Matrix<kaldi::BaseFloat>& feats,
    std::string id = "")
  : feature_info_(feature_info), mfcnet_socket_(nullptr), ex_mfcnet_socket_(nullptr), id_(id), feat_dim_(feats.NumCols()), feat_bytes_(0), num_frames_(feats.NumRows()), input_finished_(false)
{
  {
    this->source_info_.source_id = std::stoi(id);
    this->source_info_.azimuth = 0.0f;
    this->source_info_.elevation = 0.0f;
    this->source_info_.tv_sec = 0;
    this->source_info_.tv_usec = 0;
  }

  //DEBUG//{
  //DEBUG//  std::stringstream ss;
  //DEBUG//  ss << std::fixed << std::setprecision(6)
  //DEBUG//     << "source_id = "   << this->source_info_.source_id
  //DEBUG//     << ", azimuth = "   << this->source_info_.azimuth
  //DEBUG//     << ", elevation = " << this->source_info_.elevation
  //DEBUG//     << ", sec = "       << this->source_info_.tv_sec
  //DEBUG//     << ", usec = "      << this->source_info_.tv_usec;
  //DEBUG//  _raw_print_with_endl(ss.str());
  //DEBUG//}

  {
    this->features_ = feats;
    this->input_finished_ = true;
    //KALDI_LOG << "dim = " << feat_dim_;
  }

  if(this->feature_info_->feature_type.empty()){
    this->frame_shift_in_seconds_ = 0.01f;
  }
  else if(this->feature_info_->feature_type == "mfcc"){
    this->frame_shift_in_seconds_ = this->feature_info_->mfcc_opts.frame_opts.frame_shift_ms / 1000.0f;
  }
  else if(this->feature_info_->feature_type == "plp"){
    this->frame_shift_in_seconds_ = this->feature_info_->plp_opts.frame_opts.frame_shift_ms / 1000.0f;
  }
  else if(this->feature_info_->feature_type == "fbank"){
    this->frame_shift_in_seconds_ = this->feature_info_->fbank_opts.frame_opts.frame_shift_ms / 1000.0f;
  }
  else{
    this->frame_shift_in_seconds_ = 0.01f; // unknown feature type uses default value.
 }

}


OnlineFeatureMfcnet::OnlineFeatureMfcnet(
    //kaldi::OnlineFeatureInputSettingOptions &opts,
    kaldi::OnlineNnet2FeaturePipelineInfo *feature_info,
    std::shared_ptr<mfcnet_server> socket,
    std::string id = "")
  : feature_info_(feature_info), mfcnet_socket_(socket), ex_mfcnet_socket_(nullptr), id_(id), feat_dim_(0), feat_bytes_(0), num_frames_(0), input_finished_(false)
{
  if(this->mfcnet_socket_.get() == nullptr){ return; }

  do{
    this->source_info_ = this->mfcnet_socket_->buffers[this->id_]->get_source();
    if(this->mfcnet_socket_->buffers[this->id_]->srcinfo_is_ready()){ break; } // The SorceInfo has been prepared, so it exit.
    multiplatform::time::ms_sleep(10); // It wait only when a retry occurs.
    _warning_print("srcinfo is not ready");
  } while(true);

  //DEBUG//{
  //DEBUG//  std::stringstream ss;
  //DEBUG//  ss << std::fixed << std::setprecision(6)
  //DEBUG//     << "source_id = "   << this->source_info_.source_id
  //DEBUG//     << ", azimuth = "   << this->source_info_.azimuth
  //DEBUG//     << ", elevation = " << this->source_info_.elevation
  //DEBUG//     << ", sec = "       << this->source_info_.tv_sec
  //DEBUG//     << ", usec = "      << this->source_info_.tv_usec;
  //DEBUG//  _raw_print_with_endl(ss.str());
  //DEBUG//}

  do{
    // The purpose is achieved by only updating the number of frames that can be acquired,
    // so the features are discarded.
    std::vector<float> features = this->mfcnet_socket_->buffers[this->id_]->get_feature();

    // Normal case:
    // The available features exist, so it exit.
    if(this->mfcnet_socket_->buffers[this->id_]->srcdata_is_ready()){ break; }

    // Error case:
    // Even if an available feature does not exist, reception processing can not be continued, so it exit.
    if(this->mfcnet_socket_->buffers[this->id_]->is_completed()){
      KALDI_WARN << "The first frame received, but it has requested terminate.";
      this->input_finished_ = true;
      break;
      // If you need safe coding, you can use following.
      //std::int32_t available_frames = this->mfcnet_socket_->buffers[this->id_]->get_available_frames();
      //if(available_frames == 0){
      //  KALDI_WARN << "The first frame received, but it has requested terminate.";
      //  this->input_finished_ = true;
      //}
      //else{
      //  // If you reached else's route, it is a bug.
      //  // The case of reaching else's route is only when another thread tries to process data of the same client.
      //}
      //break;
    }
    multiplatform::time::ms_sleep(10); // It wait only when a retry occurs.
    _warning_print("srcdata is not ready");
  } while(true);

  {
    this->feat_bytes_ = this->mfcnet_socket_->buffers[this->id_]->get_feature_bytes();
    this->feat_dim_ = this->mfcnet_socket_->buffers[this->id_]->get_feature_dimension();
    //KALDI_LOG << "dim = " << feat_dim_;
    //CAUTION// Please be aware that if you set to true the input_finish_ flag here, the features will not be set.
    //CAUTION// if(this->mfcnet_socket_->buffers[this->id_]->is_completed()){ }
  }

  if(this->feature_info_->feature_type.empty()){
    this->frame_shift_in_seconds_ = 0.01f;
  }
  else if(this->feature_info_->feature_type == "mfcc"){
    this->frame_shift_in_seconds_ = this->feature_info_->mfcc_opts.frame_opts.frame_shift_ms / 1000.0f;
  }
  else if(this->feature_info_->feature_type == "plp"){
    this->frame_shift_in_seconds_ = this->feature_info_->plp_opts.frame_opts.frame_shift_ms / 1000.0f;
  }
  else if(this->feature_info_->feature_type == "fbank"){
    this->frame_shift_in_seconds_ = this->feature_info_->fbank_opts.frame_opts.frame_shift_ms / 1000.0f;
  }
  else{
    this->frame_shift_in_seconds_ = 0.01f; // unknown feature type uses default value.
 }

}


OnlineFeatureMfcnet::OnlineFeatureMfcnet(
    //kaldi::OnlineFeatureInputSettingOptions &opts,
    kaldi::OnlineNnet2FeaturePipelineInfo *feature_info,
    std::shared_ptr<ex_mfcnet_server> socket,
    std::string id = "")
  : feature_info_(feature_info), mfcnet_socket_(nullptr), ex_mfcnet_socket_(socket), id_(id), feat_dim_(0), feat_bytes_(0), num_frames_(0), input_finished_(false)
{
  if(this->ex_mfcnet_socket_.get() == nullptr){ return; }

  do{
    this->utterance_info_ = this->ex_mfcnet_socket_->buffers[this->id_]->get_source();
    if(this->ex_mfcnet_socket_->buffers[this->id_]->uttinfo_is_ready()){ break; } // The SorceInfo has been prepared, so it exit.
    multiplatform::time::ms_sleep(10); // It wait only when a retry occurs.
    _warning_print("utteranceinfo is not ready");
  } while(true);

  //DEBUG//{
  //DEBUG//  std::stringstream ss;
  //DEBUG//  ss << std::fixed << std::setprecision(6)
  //DEBUG//     << "client_id = "   << this->utterance_info_.client_id
  //DEBUG//     << ", mic_index = " << this->utterance_info_.mic_index
  //DEBUG//     << ", source_id = " << this->utterance_info_.source_id
  //DEBUG//     << ", sub_id = "    << this->utterance_info_.sub_id
  //DEBUG//     << ", azimuth = "   << this->utterance_info_.azimuth
  //DEBUG//     << ", elevation = " << this->utterance_info_.elevation
  //DEBUG//     << ", sec = "       << this->utterance_info_.tv_sec
  //DEBUG//     << ", usec = "      << this->utterance_info_.tv_usec;
  //DEBUG//  _raw_print_with_endl(ss.str());
  //DEBUG//}

  do{
    // The purpose is achieved by only updating the number of frames that can be acquired,
    // so the features are discarded.
    std::vector<float> features = this->ex_mfcnet_socket_->buffers[this->id_]->get_feature();

    // Normal case:
    // The available features exist, so it exit.
    if(this->ex_mfcnet_socket_->buffers[this->id_]->uttdata_is_ready()){ break; }

    // Error case:
    // Even if an available feature does not exist, reception processing can not be continued, so it exit.
    if(this->ex_mfcnet_socket_->buffers[this->id_]->is_completed()){
      KALDI_WARN << "The first frame received, but it has requested terminate.";
      this->input_finished_ = true;
      break;
      // If you need safe coding, you can use following.
      //std::int32_t available_frames = this->ex_mfcnet_socket_->buffers[this->id_]->get_available_frames();
      //if(available_frames == 0){
      //  KALDI_WARN << "The first frame received, but it has requested terminate.";
      //  this->input_finished_ = true;
      //}
      //else{
      //  // If you reached else's route, it is a bug.
      //  // The case of reaching else's route is only when another thread tries to process data of the same client.
      //}
      //break;
    }
    multiplatform::time::ms_sleep(10); // It wait only when a retry occurs.
    _warning_print("srcdata is not ready");
  } while(true);

  {
    this->feat_bytes_ = this->ex_mfcnet_socket_->buffers[this->id_]->get_feature_bytes();
    this->feat_dim_ = this->ex_mfcnet_socket_->buffers[this->id_]->get_feature_dimension();
    //KALDI_LOG << "dim = " << this->feat_dim_;
    //CAUTION// Please be aware that if you set to true the input_finish_ flag here, the features will not be set.
    //CAUTION// if(this->ex_mfcnet_socket_->buffers[this->id_]->is_completed()){ }
  }

  if(this->feature_info_->feature_type.empty()){
    this->frame_shift_in_seconds_ = 0.01f;
  }
  else if(this->feature_info_->feature_type == "mfcc"){
    this->frame_shift_in_seconds_ = this->feature_info_->mfcc_opts.frame_opts.frame_shift_ms / 1000.0f;
  }
  else if(this->feature_info_->feature_type == "plp"){
    this->frame_shift_in_seconds_ = this->feature_info_->plp_opts.frame_opts.frame_shift_ms / 1000.0f;
  }
  else if(this->feature_info_->feature_type == "fbank"){
    this->frame_shift_in_seconds_ = this->feature_info_->fbank_opts.frame_opts.frame_shift_ms / 1000.0f;
  }
  else{
    this->frame_shift_in_seconds_ = 0.01f; // unknown feature type uses default value.
 }

}

OnlineFeatureMfcnet::~OnlineFeatureMfcnet() {
  //DEBUG//KALDI_LOG << "Destructor of " << this->id_ << " finished.\n";
  if(this->mfcnet_socket_.get() != nullptr){
    this->mfcnet_socket_->erase_buffer_key(this->id_);
    this->mfcnet_socket_->erase_terminated_key(this->id_);
    this->mfcnet_socket_.reset();
  }
  if(this->ex_mfcnet_socket_.get() != nullptr){
    this->ex_mfcnet_socket_->erase_buffer_key(this->id_);
    this->ex_mfcnet_socket_->erase_terminated_key(this->id_);
    this->ex_mfcnet_socket_.reset();
  }
}

void OnlineFeatureMfcnet::Initialize() {
}

kaldi::int32 OnlineFeatureMfcnet::Dim() const
{
  return this->feat_dim_;
}

kaldi::int32 OnlineFeatureMfcnet::NumFramesReady() const
{
  return this->num_frames_;
}

bool OnlineFeatureMfcnet::IsLastFrame(kaldi::int32 frame) const
{
  //DEBUG//if(input_finished_){ KALDI_WARN << "\nid=" << this->id_ << ", frame=" << frame << ", this->num_frames_=" << this->num_frames_ << ", this->features_.NumRow()=" << this->features_.NumRows(); }
  return (this->input_finished_ ? ((frame + 1) == this->num_frames_) : false);
}

void OnlineFeatureMfcnet::GetFrame(kaldi::int32 frame, kaldi::VectorBase<kaldi::BaseFloat> *feat)
{
  //this->Receive();
  KALDI_ASSERT(frame >= 0 && frame < this->num_frames_);
  KALDI_ASSERT(feat->Dim() == this->Dim());
  feat->CopyFromVec(this->features_.Row(frame));
}

void OnlineFeatureMfcnet::MfcnetReceive()
{ // feat_dim_ == Dim()
  //DEBUG//KALDI_WARN << "id=" << this->id_ << "mfcnet_socket_==nullptr? : " << (this->mfcnet_socket_.get() == nullptr);
  //DEBUG//KALDI_WARN << "id=" << this->id_ << "ex_mfcnet_socket_==nullptr? : " << (this->ex_mfcnet_socket_.get() == nullptr);
  if(this->mfcnet_socket_.get() == nullptr){ return; }
  //DEBUG//KALDI_WARN << "id=" << this->id_ << "input_finished_? : " << input_finished_;
  if(this->input_finished_ == true){ return; }

  if(this->mfcnet_socket_->buffers[this->id_]->is_completed()){
    //DEBUG//KALDI_WARN << "id=" << this->id_ << " received all data based on the mfcnet protocol.";
    this->input_finished_ = true;
  }
  else{
    // Since the socket was disconnected without receiving the termination packet of the feature,
    // an warning message is output and a completion flag is set.
    if(this->mfcnet_socket_->find_terminated_key(this->id_)){
      //DEBUG//KALDI_WARN << "id=" << this->id_ << " detected disconnection of the socket before receiving all data.";
      this->input_finished_ = true;
    }
  }

  std::vector<float> feat_buf = this->mfcnet_socket_->buffers[this->id_]->get_feature();
  //std::vector<float> mask_buf = this->mfcnet_socket_->buffers[this->id_]->get_mask(); // This is dummy function.
  std::int32_t available_frames = this->mfcnet_socket_->buffers[this->id_]->get_available_frames();
  //DEBUG//KALDI_WARN << "id=" << this->id_ << "available_frames=" << available_frames << ", this->num_frames_=" << this->num_frames_;

  // You can trust the results obtained with get_available_frames() unless get_feature() is executed.
  // If you know that multiple threads do not work for the same client session (same utterance),
  // the following checks are no need.
  //if(feat_buf.size() < (available_frames * this->feat_dim_)){
  //  available_frames = (feat_buf.size() / this->feat_dim_);
  //}
  if(available_frames > num_frames_){
    kaldi::Matrix<kaldi::BaseFloat> feats;
    //kaldi::Matrix<kaldi::BaseFloat> masks;

#if 1 // This is a method to preparing a new buffer and then swapping it with the old buffer.
    // This is initialization of Matrix, so kCopyData is not needed as the 3rd argument of Resize method.
    feats.Resize(available_frames, this->feat_dim_);
    //masks.Resize(available_frames, this->feat_dim_);

    for(int i=0; i<available_frames; i++){
      std::memcpy(feats.RowData(i), &feat_buf.at(this->feat_dim_*i), this->feat_bytes_);
      //std::memcpy(masks.RowData(i), mask_buf.at(this->feat_dim_*i), this->feat_bytes_);
    }

    if(available_frames != feats.NumRows()){
      KALDI_WARN << "available_frames=" << available_frames << ", feats.NumRows()=" << feats.NumRows();
    }

    KALDI_VLOG(3) << "source_id = " << this->source_info_.source_id
                  << ", received frames = " << available_frames;

    this->features_.Swap(&feats);
    feats.Resize(0, 0);
    malloc_trim(0);
    this->num_frames_ = available_frames;
#else // This is a buffer resizing and joining method similar to conventional behavior.
    std::int32_t increased_frames = available_frames - this->num_frames_;
    // This is initialization of Matrix, so kCopyData is not needed as the 3rd argument of Resize method.
    feats.Resize(increased_frames, this->feat_dim_);
    //masks.Resize(increased_frames, this->feat_dim_);

    for(int i=0, j=this->num_frames_; i<increased_frames; i++, j++){
      std::memcpy(feats.RowData(i), &feat_buf.at(this->feat_dim_*j), this->feat_bytes_);
      //std::memcpy(masks.RowData(i), mask_buf.at(this->feat_dim_*j), this->feat_bytes_);
    }

    if(increased_frames != feats.NumRows()){
      KALDI_WARN << "increased_frames=" << increased_frames << ", feats.NumRows()=" << feats.NumRows();
    }

    KALDI_VLOG(3) << "source_id = " << this->source_info_.source_id
                  << ", received frames = " << available_frames;

    std::int32_t new_num_frames = this->num_frames_ + feats.NumRows(); // feats.NumRows() is increased_frames.
    const float increase_ratio = 1.5f;
    if(new_num_frames > this->features_.NumRows()){
      kaldi::int32 new_num_rows = std::max<kaldi::int32>(new_num_frames, this->features_.NumRows() * increase_ratio);
      this->features_.Resize(new_num_rows, this->feat_dim_, kCopyData);
    }
    this->features_.Range(this->num_frames_, feats.NumRows(), 0, this->feat_dim_).CopyFromMat(feats);
    this->num_frames_ = new_num_frames;
#endif // SELECT_***_MODE
  }

}

void OnlineFeatureMfcnet::ExMfcnetReceive()
{ // feat_dim_ == this->Dim()
  //DEBUG//KALDI_WARN << "id=" << this->id_ << "mfcnet_socket_==nullptr? : " << (this->mfcnet_socket_.get() == nullptr);
  //DEBUG//KALDI_WARN << "id=" << this->id_ << "ex_mfcnet_socket_==nullptr? : " << (this->ex_mfcnet_socket_.get() == nullptr);
  if(this->ex_mfcnet_socket_.get() == nullptr){ return; }
  //DEBUG//KALDI_WARN << "id=" << this->id_ << "input_finished_? : " << this->input_finished_;
  if(this->input_finished_ == true){ return; }

  if(this->ex_mfcnet_socket_->buffers[this->id_]->is_completed()){
    //DEBUG//KALDI_WARN << "id=" << this->id_ << " received all data based on the mfcnet protocol.";
    this->input_finished_ = true;
  }
  else{
    // Since the socket was disconnected without receiving the termination packet of the feature,
    // an warning message is output and a completion flag is set.
    if(this->ex_mfcnet_socket_->find_terminated_key(this->id_)){
      //DEBUG//KALDI_WARN << "id=" << this->id_ << " detected disconnection of the socket before receiving all data.";
      this->input_finished_ = true;
    }
  }

  std::vector<float> feat_buf = this->ex_mfcnet_socket_->buffers[this->id_]->get_feature();
  //std::vector<float> mask_buf = this->ex_mfcnet_socket_->buffers[this->id_]->get_mask(); // This is dummy function.
  std::int32_t available_frames = this->ex_mfcnet_socket_->buffers[this->id_]->get_available_frames();
  //DEBUG//KALDI_WARN << "id=" << this->id_ << "available_frames=" << available_frames << ", this->num_frames_=" << this->num_frames_;

  // You can trust the results obtained with get_available_frames() unless get_feature() is executed.
  // If you know that multiple threads do not work for the same client session (same utterance),
  // the following checks are no need.
  //if(feat_buf.size() < (available_frames * this->feat_dim_)){
  //  available_frames = (feat_buf.size() / this->feat_dim_);
  //}
  if(available_frames > this->num_frames_){
    kaldi::Matrix<kaldi::BaseFloat> feats;
    //kaldi::Matrix<kaldi::BaseFloat> masks;

#if 1 // This is a method to preparing a new buffer and then swapping it with the old buffer.
    // This is initialization of Matrix, so kCopyData is not needed as the 3rd argument of Resize method.
    feats.Resize(available_frames, this->feat_dim_);
    //masks.Resize(available_frames, this->feat_dim_);

    for(int i=0; i<available_frames; i++){
      std::memcpy(feats.RowData(i), &feat_buf.at(this->feat_dim_*i), this->feat_bytes_);
      //std::memcpy(masks.RowData(i), mask_buf.at(this->feat_dim_*i), this->feat_bytes_);
    }

    if(available_frames != feats.NumRows()){
      KALDI_WARN << "available_frames=" << available_frames << ", feats.NumRows()=" << feats.NumRows();
    }

    KALDI_VLOG(3) << "source_id = " << this->utterance_info_.source_id
                  << ", received frames = " << available_frames;

    this->features_.Swap(&feats);
    feats.Resize(0, 0);
    malloc_trim(0);
    this->num_frames_ = available_frames;
#else // This is a buffer resizing and joining method similar to conventional behavior.
    std::int32_t increased_frames = available_frames - this->num_frames_;
    // This is initialization of Matrix, so kCopyData is not needed as the 3rd argument of Resize method.
    feats.Resize(increased_frames, this->feat_dim_);
    //masks.Resize(increased_frames, this->feat_dim_);

    for(int i=0, j=num_frames_; i<increased_frames; i++, j++){
      std::memcpy(feats.RowData(i), &feat_buf.at(this->feat_dim_*j), this->feat_bytes_);
      //std::memcpy(masks.RowData(i), mask_buf.at(this->feat_dim_*j), this->feat_bytes_);
    }

    if(increased_frames != feats.NumRows()){
      KALDI_WARN << "increased_frames=" << increased_frames << ", feats.NumRows()=" << feats.NumRows();
    }

    KALDI_VLOG(3) << "source_id = " << this->utterance_info_.source_id
                  << ", received frames = " << available_frames;

    std::int32_t new_num_frames = this->num_frames_ + feats.NumRows(); // feats.NumRows() is increased_frames.
    const float increase_ratio = 1.5f;
    if(new_num_frames > features_.NumRows()){
      kaldi::int32 new_num_rows = std::max<kaldi::int32>(new_num_frames, this->features_.NumRows() * increase_ratio);
      this->features_.Resize(new_num_rows, this->feat_dim_, kCopyData);
    }
    this->features_.Range(this->num_frames_, feats.NumRows(), 0, this->feat_dim_).CopyFromMat(feats);
    this->num_frames_ = new_num_frames;
#endif // SELECT_***_MODE
  }

}

//#if (KALDI_DOUBLEPRECISION != 0) is not supported. if you use double, received socket data and goto copy feat matrix
void OnlineFeatureMfcnet::Receive()
{
  if(this->mfcnet_socket_.get() != nullptr){ this->MfcnetReceive(); }
  if(this->ex_mfcnet_socket_.get() != nullptr){ this->ExMfcnetReceive(); }
}

} // namespace kaldi
