// read_stats.hpp: Include file used to read system statistics

#ifndef MP_READ_STATS_H_
#define MP_READ_STATS_H_

namespace multiplatform {

// ***************************************************************************
// * Definitions of structures for system statistics.
// * WARNING: Fields order matters for SVG graphs!
// ***************************************************************************

// * Structure for CPU statistics.
// * In activity buffer: First structure is for global CPU utilisation ("all").
// * Following structures are for each individual CPU (0, 1, etc.)

  struct stats_cpu {
    uint64_t cpu_user;
    uint64_t cpu_nice;
    uint64_t cpu_sys;
    uint64_t cpu_idle;
    uint64_t cpu_iowait;
    uint64_t cpu_steal;
    uint64_t cpu_hardirq;
    uint64_t cpu_softirq;
    uint64_t cpu_guest;
    uint64_t cpu_guest_nice;
  };

  struct stats_memory {
    uint64_t frmkb;
    uint64_t bufkb;
    uint64_t camkb;
    uint64_t tlmkb;
    uint64_t frskb;
    uint64_t tlskb;
    uint64_t caskb;
    uint64_t comkb;
    uint64_t activekb;
    uint64_t inactkb;
    uint64_t dirtykb;
    uint64_t anonpgkb;
    uint64_t slabkb;
    uint64_t kstackkb;
    uint64_t pgtblkb;
    uint64_t vmusedkb;
    uint64_t availablekb;
  };

  //unsigned int kb_shift;

class read_statistics {
public:
/*
  struct stats_cpu {
    uint64_t cpu_user;
    uint64_t cpu_nice;
    uint64_t cpu_sys;
    uint64_t cpu_idle;
    uint64_t cpu_iowait;
    uint64_t cpu_steal;
    uint64_t cpu_hardirq;
    uint64_t cpu_softirq;
    uint64_t cpu_guest;
    uint64_t cpu_guest_nice;
  };

  struct stats_memory {
    uint64_t frmkb;
    uint64_t bufkb;
    uint64_t camkb;
    uint64_t tlmkb;
    uint64_t frskb;
    uint64_t tlskb;
    uint64_t caskb;
    uint64_t comkb;
    uint64_t activekb;
    uint64_t inactkb;
    uint64_t dirtykb;
    uint64_t anonpgkb;
    uint64_t slabkb;
    uint64_t kstackkb;
    uint64_t pgtblkb;
    uint64_t vmusedkb;
    uint64_t availablekb;
  };
*/

  uint64_t _hz;
  unsigned int _kb_shift;

  read_statistics(void){}
  ~read_statistics(void){}

  int read_stat_cpu(struct stats_cpu *, int);
  int read_meminfo(struct stats_memory *);
  void read_uptime(uint64_t *);

  // from count
  unsigned int get_cpu_nr(unsigned int, int);

  // from common
  void get_HZ(void);
  void get_kb_shift(void);

  unsigned int KB_TO_PG(uint64_t k){ return (k >> _kb_shift); }
  unsigned int PG_TO_KB(uint64_t k){ return (k << _kb_shift); }

private:
  // from count
  unsigned int get_sys_cpu_nr(int);
  unsigned int get_proc_cpu_nr(void);
};

} // namespace multiplatform

#endif // MP_RD_STATS_H_

