// statistics print base

#ifndef MP_STAT_PRINT_BASE_H_
#define MP_STAT_PRINT_BASE_H_

#include <time.h>
#include <limits.h>
#include <stdlib.h>

#include <string>

namespace multiplatform {

// ***************************************************************************
// * Various keywords and constants
// ***************************************************************************

constexpr static int FALSE = 0;
constexpr static int TRUE = 1;

constexpr static int PLAIN_OUTPUT = 0;

constexpr static int DISP_HDR = 1;

// Keywords
constexpr static char K_ISO[] = "ISO";
constexpr static char K_ALL[] = "ALL";
constexpr static char K_LOWERALL[] = "all";
constexpr static char K_UTC[] = "UTC";
constexpr static char K_JSON[] = "JSON";

// ***************************************************************************
// * Functions prototypes
// ***************************************************************************
void print_version(void);

time_t get_localtime(struct tm *, int);

uint64_t get_interval(uint64_t, uint64_t);

int get_win_height(void);

constexpr double sp_value(uint64_t value1, uint64_t value2, uint64_t itv){
  return (value2 < value1) ? (double) 0 : (((double) (value2 - value1)) / itv * 100);
}
constexpr double s_value(uint64_t value1, uint64_t value2, uint64_t itv){
  return (value2 < value1) ? (double) 0 : (((double) (value2 - value1)) / itv * 100);
}

int is_iso_time_fmt(void);
int print_gal_header(struct tm *, char *, char *, char *, char *, int, int);

void xprintf(int, const char *, ...);

class time_support {
public:

private:

};


class cprint_support {
public:
  // Index in units array (see common.c)
  constexpr static int NO_UNIT       = -1;
  constexpr static int UNIT_SECTOR   = 0;
  constexpr static int UNIT_BYTE     = 1;
  constexpr static int UNIT_KILOBYTE = 2;
  constexpr static int UNIT_MEGABYTE = 3;
  constexpr static int UNIT_GIGABYTE = 4;
  constexpr static int UNIT_TERABYTE = 5;
  constexpr static int UNIT_PETABYTE = 6;
  constexpr static int UNIT_UNKNOWN  = 7;
  constexpr static int NR_UNITS      = 8;

  // Colors definitions
  constexpr static int IS_INT     = 0;
  constexpr static int IS_STR     = 1;
  constexpr static int IS_RESTART = 2;
  constexpr static int IS_COMMENT = 3;
  constexpr static int IS_ZERO    = 4;

  cprint_support(){ init_colors(); }
  ~cprint_support(){ }

  void init_colors(void);

  void cprintf_f(int, int, int, int, ...);
  void cprintf_in(int, char *, char *, int);
  void cprintf_pc(int, int, int, int, ...);
  void cprintf_s(int, char *, char *);
  void cprintf_u64(int, int, int, ...);
  void cprintf_x(int, int, ...);

private:
  constexpr static size_t MAX_SGR_LEN = 16;

  // Colors strings
  char sc_percent_high[MAX_SGR_LEN] = "\e[31;22m"; // BOLD RED
  char sc_percent_low[MAX_SGR_LEN] = "\e[35;1m"; // BOLD MAGENTA
  char sc_zero_int_stat[MAX_SGR_LEN] = "\e[34;22m"; // LIGHT BLUE
  char sc_int_stat[MAX_SGR_LEN] = "\e[34;1m"; // BOLD BLUE
  char sc_item_name[MAX_SGR_LEN] = "\e[32;22m"; // LIGHT GREEN
  char sc_sa_restart[MAX_SGR_LEN] = "\e[31;22m"; // LIGHT RED
  char sc_sa_comment[MAX_SGR_LEN] = "\e[33;22m"; // LIGHT YELLOW
  char sc_normal[MAX_SGR_LEN] = "\e[0m"; // NORMAL

  void cprintf_unit(int, int, double);
};



} // namespace multiplatform

#endif // MP_COMMON_H_

